@@ -18,10 +18,10 @@ import ast
 import errno
 import json
 import os
-import pipes
 import platform
 import pprint
 import re
+import shlex
 import shutil
 import sys
 import subprocess
@@ -702,6 +702,8 @@ class MetaBuildWrapper():
       swarming_targets = set(contents.splitlines())
 
       isolate_map = self.ReadIsolateMap()
+      self.RemovePossiblyStaleRuntimeDepsFiles(vals, swarming_targets,
+                                               isolate_map, build_dir)
       err, labels = self.MapTargetsToLabels(isolate_map, swarming_targets)
       if err:
         raise MBErr(err)
@@ -729,6 +731,7 @@ class MetaBuildWrapper():
         label = isolate_map[target]['label']
         runtime_deps_targets = [
             target + '.runtime_deps',
+            'obj/%s.runtime_deps' % label.replace(':', '/'),
             'obj/%s.stamp.runtime_deps' % label.replace(':', '/')]
       elif (isolate_map[target]['type'] == 'script' or
             isolate_map[target].get('label_type') == 'group'):
@@ -738,6 +741,7 @@ class MetaBuildWrapper():
         # also be an executable.
         label = isolate_map[target]['label']
         runtime_deps_targets = [
+            'obj/%s.runtime_deps' % label.replace(':', '/'),
             'obj/%s.stamp.runtime_deps' % label.replace(':', '/')]
         if self.platform == 'win32':
           runtime_deps_targets += [ target + '.exe.runtime_deps' ]
@@ -793,6 +797,99 @@ class MetaBuildWrapper():
 
     return ret
 
+  def RemovePossiblyStaleRuntimeDepsFiles(self, vals, targets, isolate_map,
+                                          build_dir):
+    # TODO(crbug.com/41441724): Because `gn gen --runtime-deps-list-file`
+    # puts the runtime_deps file in different locations based on the actual
+    # type of a target, we may end up with multiple possible runtime_deps
+    # files in a given build directory, where some of the entries might be
+    # stale (since we might be reusing an existing build directory).
+    #
+    # We need to be able to get the right one reliably; you might think
+    # we can just pick the newest file, but because GN won't update timestamps
+    # if the contents of the files change, an older runtime_deps
+    # file might actually be the one we should use over a newer one (see
+    # crbug.com/932387 for a more complete explanation and example).
+    #
+    # In order to avoid this, we need to delete any possible runtime_deps
+    # files *prior* to running GN. As long as the files aren't actually
+    # needed during the build, this hopefully will not cause unnecessary
+    # build work, and so it should be safe.
+    #
+    # Ultimately, we should just make sure we get the runtime_deps files
+    # in predictable locations so we don't have this issue at all, and
+    # that's what crbug.com/932700 is for.
+    possible_rpaths = self.PossibleRuntimeDepsPaths(vals, targets, isolate_map)
+    for rpaths in possible_rpaths.values():
+      for rpath in rpaths:
+        path = self.ToAbsPath(build_dir, rpath)
+        if self.Exists(path):
+          self.RemoveFile(path)
+
+  def PossibleRuntimeDepsPaths(self, vals, ninja_targets, isolate_map):
+    """Returns a map of targets to possible .runtime_deps paths.
+
+    Each ninja target maps on to a GN label, but depending on the type
+    of the GN target, `gn gen --runtime-deps-list-file` will write
+    the .runtime_deps files into different locations. Unfortunately, in
+    some cases we don't actually know which of multiple locations will
+    actually be used, so we return all plausible candidates.
+
+    The paths that are returned are relative to the build directory.
+    """
+
+    android = 'target_os="android"' in vals['gn_args']
+    ios = 'target_os="ios"' in vals['gn_args']
+    fuchsia = 'target_os="fuchsia"' in vals['gn_args']
+    win = self.platform == 'win32' or 'target_os="win"' in vals['gn_args']
+    possible_runtime_deps_rpaths = {}
+    for target in ninja_targets:
+      target_type = isolate_map[target]['type']
+      label = isolate_map[target]['label']
+      target_runtime_deps = 'obj/%s.runtime_deps' % label.replace(':', '/')
+      stamp_runtime_deps = 'obj/%s.stamp.runtime_deps' % label.replace(':', '/')
+      # TODO(crbug.com/40590196): 'official_tests' use
+      # type='additional_compile_target' to isolate tests. This is not the
+      # intended use for 'additional_compile_target'.
+      if (target_type == 'additional_compile_target' and
+          target != 'official_tests'):
+        # By definition, additional_compile_targets are not tests, so we
+        # shouldn't generate isolates for them.
+        raise MBErr('Cannot generate isolate for %s since it is an '
+                    'additional_compile_target.' % target)
+      if fuchsia or ios or target_type == 'generated_script':
+        # iOS and Fuchsia targets end up as groups.
+        # generated_script targets are always actions.
+        rpaths = [stamp_runtime_deps, target_runtime_deps]
+      elif android:
+        # Android targets may be either android_apk or executable. The former
+        # will result in runtime_deps associated with the stamp file, while the
+        # latter will result in runtime_deps associated with the executable.
+        label = isolate_map[target]['label']
+        rpaths = [
+            target + '.runtime_deps', stamp_runtime_deps, target_runtime_deps
+        ]
+      elif (target_type == 'script' or
+            isolate_map[target].get('label_type') == 'group'):
+        # For script targets, the build target is usually a group,
+        # for which gn generates the runtime_deps next to the stamp file
+        # for the label, which lives under the obj/ directory, but it may
+        # also be an executable.
+        label = isolate_map[target]['label']
+        rpaths = [stamp_runtime_deps, target_runtime_deps]
+        if win:
+          rpaths += [target + '.exe.runtime_deps']
+        else:
+          rpaths += [target + '.runtime_deps']
+      elif win:
+        rpaths = [target + '.exe.runtime_deps']
+      else:
+        rpaths = [target + '.runtime_deps']
+
+      possible_runtime_deps_rpaths[target] = rpaths
+
+    return possible_runtime_deps_rpaths
+
   def WriteIsolateFiles(self, build_dir, target, runtime_deps):
     isolate_path = self.ToAbsPath(build_dir, target + '.isolate')
     self.WriteFile(isolate_path,
@@ -1100,8 +1197,8 @@ class MetaBuildWrapper():
       shell_quoter = QuoteForCmd
     else:
       env_prefix = ''
-      env_quoter = pipes.quote
-      shell_quoter = pipes.quote
+      env_quoter = shlex.quote
+      shell_quoter = shlex.quote
 
     def print_env(var):
       if env and var in env:
