@@ -802,7 +802,7 @@ Socket.prototype[SymbolAsyncDispose] = async function() {
   if (!this[kStateSymbol].handle) {
     return;
   }
-  return FunctionPrototypeCall(promisify(this.close), this);
+  await FunctionPrototypeCall(promisify(this.close), this);
 };
 
 
