@@ -21,7 +21,6 @@ const {
   refreshOptions,
   getEmbedderOptions,
 } = require('internal/options');
-const { reconnectZeroFillToggle } = require('internal/buffer');
 const {
   exposeLazyInterfaces,
   defineReplaceableLazyAttribute,
@@ -51,6 +50,15 @@ function prepareMainThreadExecution(expandArgv1 = false, initializeModules = tru
   });
 }
 
+function prepareTestRunnerMainExecution(loadUserModules = true) {
+  return prepareExecution({
+    expandArgv1: false,
+    initializeModules: true,
+    isMainThread: true,
+    forceDefaultLoader: !loadUserModules,
+  });
+}
+
 function prepareWorkerThreadExecution() {
   prepareExecution({
     expandArgv1: false,
@@ -88,10 +96,9 @@ function prepareShadowRealmExecution() {
 }
 
 function prepareExecution(options) {
-  const { expandArgv1, initializeModules, isMainThread } = options;
+  const { expandArgv1, initializeModules, isMainThread, forceDefaultLoader } = options;
 
   refreshRuntimeOptions();
-  reconnectZeroFillToggle();
 
   // Patch the process object and get the resolved main entry point.
   const mainEntry = patchProcessObject(expandArgv1);
@@ -101,6 +108,7 @@ function prepareExecution(options) {
   setupNavigator();
   setupWarningHandler();
   setupSQLite();
+  setupQuic();
   setupWebStorage();
   setupWebsocket();
   setupEventsource();
@@ -115,6 +123,8 @@ function prepareExecution(options) {
   initializeSourceMapsHandlers();
   initializeDeprecations();
 
+  initializeConfigFileSupport();
+
   require('internal/dns/utils').initializeDns();
 
   if (isMainThread) {
@@ -145,12 +155,37 @@ function prepareExecution(options) {
   }
 
   if (initializeModules) {
-    setupUserModules();
+    setupUserModules(forceDefaultLoader);
   }
 
+  // This has to be done after the user module loader is initialized,
+  // in case undici is externalized.
+  setupHttpProxy();
+
   return mainEntry;
 }
 
+function setupHttpProxy() {
+  // This normalized from both --use-env-proxy and NODE_USE_ENV_PROXY settings.
+  if (!getOptionValue('--use-env-proxy')) {
+    return;
+  }
+  if (!process.env.HTTP_PROXY && !process.env.HTTPS_PROXY &&
+      !process.env.http_proxy && !process.env.https_proxy) {
+    return;
+  }
+
+  const { setGlobalDispatcher, EnvHttpProxyAgent } = require('internal/deps/undici/undici');
+  const envHttpProxyAgent = new EnvHttpProxyAgent();
+  setGlobalDispatcher(envHttpProxyAgent);
+  // For fetch, we need to set the global dispatcher from here.
+  // For http/https agents, we'll configure the global agent when they are
+  // actually created, in lib/_http_agent.js and lib/https.js.
+  // TODO(joyeecheung): This is currently guarded with NODE_USE_ENV_PROXY and --use-env-proxy.
+  // Investigate whether it's possible to enable it by default without stepping on other
+  // existing libraries that sets the global dispatcher or monkey patches the global agent.
+}
+
 function setupUserModules(forceDefaultLoader = false) {
   initializeCJSLoader();
   initializeESMLoader(forceDefaultLoader);
@@ -182,7 +217,8 @@ function refreshRuntimeOptions() {
  * Replace `process.argv[0]` with `process.execPath`, preserving the original `argv[0]` value as `process.argv0`.
  * Replace `process.argv[1]` with the resolved absolute file path of the entry point, if found.
  * @param {boolean} expandArgv1 - Whether to replace `process.argv[1]` with the resolved absolute file path of
- * the main entry point.
+ *   the main entry point.
+ * @returns {string}
  */
 function patchProcessObject(expandArgv1) {
   const binding = internalBinding('process_methods');
@@ -311,6 +347,22 @@ function setupSQLite() {
   BuiltinModule.allowRequireByUsers('sqlite');
 }
 
+function initializeConfigFileSupport() {
+  if (getOptionValue('--experimental-default-config-file') ||
+      getOptionValue('--experimental-config-file')) {
+    emitExperimentalWarning('--experimental-config-file');
+  }
+}
+
+function setupQuic() {
+  if (!getOptionValue('--experimental-quic')) {
+    return;
+  }
+
+  const { BuiltinModule } = require('internal/bootstrap/realm');
+  BuiltinModule.allowRequireByUsers('quic');
+}
+
 function setupWebStorage() {
   if (getEmbedderOptions().noBrowserGlobals ||
       !getOptionValue('--experimental-webstorage')) {
@@ -341,10 +393,7 @@ function setupStacktracePrinterOnSigint() {
   if (!getOptionValue('--trace-sigint')) {
     return;
   }
-  const { SigintWatchdog } = require('internal/watchdog');
-
-  const watchdog = new SigintWatchdog();
-  watchdog.start();
+  require('internal/util/trace_sigint').setTraceSigInt(true);
 }
 
 function initializeReport() {
@@ -567,15 +616,8 @@ function initializePermission() {
       },
     });
   } else {
-    const availablePermissionFlags = [
-      '--allow-fs-read',
-      '--allow-fs-write',
-      '--allow-addons',
-      '--allow-child-process',
-      '--allow-wasi',
-      '--allow-worker',
-    ];
-    ArrayPrototypeForEach(availablePermissionFlags, (flag) => {
+    const { availableFlags } = require('internal/process/permission');
+    ArrayPrototypeForEach(availableFlags(), (flag) => {
       const value = getOptionValue(flag);
       if (value.length) {
         throw new ERR_MISSING_OPTION('--permission');
@@ -608,9 +650,17 @@ function initializeESMLoader(forceDefaultLoader) {
 
 function initializeSourceMapsHandlers() {
   const {
-    setSourceMapsEnabled,
+    setSourceMapsSupport,
   } = require('internal/source_map/source_map_cache');
-  setSourceMapsEnabled(getOptionValue('--enable-source-maps'));
+  const enabled = getOptionValue('--enable-source-maps');
+  setSourceMapsSupport(enabled, {
+    __proto__: null,
+    // TODO(legendecas): In order to smoothly improve the source map support,
+    // skip source maps in node_modules and generated code with
+    // `--enable-source-maps` in a semver major version.
+    nodeModules: enabled,
+    generatedCode: enabled,
+  });
 }
 
 function initializeFrozenIntrinsics() {
@@ -671,6 +721,7 @@ module.exports = {
   prepareMainThreadExecution,
   prepareWorkerThreadExecution,
   prepareShadowRealmExecution,
+  prepareTestRunnerMainExecution,
   markBootstrapComplete,
   loadPreloadModules,
   initializeFrozenIntrinsics,
