@@ -1,6 +1,8 @@
 'use strict';
 
 const {
+  AggregateError,
+  AggregateErrorPrototype,
   Array,
   ArrayBuffer,
   ArrayBufferPrototype,
@@ -74,6 +76,8 @@ const {
   ObjectSetPrototypeOf,
   Promise,
   PromisePrototype,
+  RangeError,
+  RangeErrorPrototype,
   ReflectApply,
   ReflectOwnKeys,
   RegExp,
@@ -114,6 +118,8 @@ const {
   SymbolPrototypeValueOf,
   SymbolToPrimitive,
   SymbolToStringTag,
+  TypeError,
+  TypeErrorPrototype,
   TypedArray,
   TypedArrayPrototype,
   TypedArrayPrototypeGetLength,
@@ -244,7 +250,6 @@ const keyStrRegExp = /^[a-zA-Z_][a-zA-Z_0-9]*$/;
 const numberRegExp = /^(0|[1-9][0-9]*)$/;
 
 const coreModuleRegExp = /^ {4}at (?:[^/\\(]+ \(|)node:(.+):\d+:\d+\)?$/;
-const nodeModulesRegExp = /[/\\]node_modules[/\\](.+?)(?=[/\\])/g;
 
 const classRegExp = /^(\s+[^(]*?)\s*{/;
 // eslint-disable-next-line node-core/no-unescaped-regexp-dot
@@ -636,7 +641,12 @@ const wellKnownPrototypes = new SafeMap()
   .set(RegExpPrototype, { name: 'RegExp', constructor: RegExp })
   .set(DatePrototype, { name: 'Date', constructor: Date })
   .set(DataViewPrototype, { name: 'DataView', constructor: DataView })
+
   .set(ErrorPrototype, { name: 'Error', constructor: Error })
+  .set(AggregateErrorPrototype, { name: 'AggregateError', constructor: AggregateError })
+  .set(RangeErrorPrototype, { name: 'RangeError', constructor: RangeError })
+  .set(TypeErrorPrototype, { name: 'TypeError', constructor: TypeError })
+
   .set(BooleanPrototype, { name: 'Boolean', constructor: Boolean })
   .set(NumberPrototype, { name: 'Number', constructor: Number })
   .set(StringPrototype, { name: 'String', constructor: String })
@@ -1305,13 +1315,122 @@ function identicalSequenceRange(a, b) {
           len++;
         }
         if (len > 3) {
-          return { len, offset: i };
+          return [len, i];
         }
       }
     }
   }
 
-  return { len: 0, offset: 0 };
+  return [0, 0];
+}
+
+function getDuplicateErrorFrameRanges(frames) {
+  // Build a map: frame line -> sorted list of indices where it occurs
+  const result = [];
+  const lineToPositions = new SafeMap();
+
+  for (let i = 0; i < frames.length; i++) {
+    const positions = lineToPositions.get(frames[i]);
+    if (positions === undefined) {
+      lineToPositions.set(frames[i], [i]);
+    } else {
+      positions[positions.length] = i;
+    }
+  }
+
+  const minimumDuplicateRange = 3;
+  // Not enough duplicate lines to consider collapsing
+  if (frames.length - lineToPositions.size <= minimumDuplicateRange) {
+    return result;
+  }
+
+  for (let i = 0; i < frames.length - minimumDuplicateRange; i++) {
+    const positions = lineToPositions.get(frames[i]);
+    // Find the next occurrence of the same line after i, if any
+    if (positions.length === 1 || positions[positions.length - 1] === i) {
+      continue;
+    }
+
+    const current = positions.indexOf(i) + 1;
+    if (current === positions.length) {
+      continue;
+    }
+
+    // Theoretical maximum range, adjusted while iterating
+    let range = positions[positions.length - 1] - i;
+    if (range < minimumDuplicateRange) {
+      continue;
+    }
+    let extraSteps;
+    if (current + 1 < positions.length) {
+      // Optimize initial step size by choosing the greatest common divisor (GCD)
+      // of all candidate distances to the same frame line. This tends to match
+      // the true repeating block size and minimizes fallback iterations.
+      let gcdRange = 0;
+      for (let j = current; j < positions.length; j++) {
+        let distance = positions[j] - i;
+        while (distance !== 0) {
+          const remainder = gcdRange % distance;
+          if (gcdRange !== 0) {
+            // Add other possible ranges as fallback
+            extraSteps ??= new SafeSet();
+            extraSteps.add(gcdRange);
+          }
+          gcdRange = distance;
+          distance = remainder;
+        }
+        if (gcdRange === 1) break;
+      }
+      range = gcdRange;
+      if (extraSteps) {
+        extraSteps.delete(range);
+        extraSteps = [...extraSteps];
+      }
+    }
+    let maxRange = range;
+    let maxDuplicates = 0;
+
+    let duplicateRanges = 0;
+
+    for (let nextStart = i + range; /* ignored */ ; nextStart += range) {
+      let equalFrames = 0;
+      for (let j = 0; j < range; j++) {
+        if (frames[i + j] !== frames[nextStart + j]) {
+          break;
+        }
+        equalFrames++;
+      }
+      // Adjust the range to match different type of ranges.
+      if (equalFrames !== range) {
+        if (!extraSteps?.length) {
+          break;
+        }
+        // Memorize former range in case the smaller one would hide less.
+        if (duplicateRanges !== 0 && maxRange * maxDuplicates < range * duplicateRanges) {
+          maxRange = range;
+          maxDuplicates = duplicateRanges;
+        }
+        range = extraSteps.pop();
+        nextStart = i;
+        duplicateRanges = 0;
+        continue;
+      }
+      duplicateRanges++;
+    }
+
+    if (maxDuplicates !== 0 && maxRange * maxDuplicates >= range * duplicateRanges) {
+      range = maxRange;
+      duplicateRanges = maxDuplicates;
+    }
+
+    if (duplicateRanges * range >= 3) {
+      result.push(i + range, range, duplicateRanges);
+      // Skip over the collapsed portion to avoid overlapping matches.
+      i += range * (duplicateRanges + 1) - 1;
+    }
+  }
+
+  return result;
 }
 
 function getStackString(ctx, error) {
@@ -1345,7 +1464,7 @@ function getStackFrames(ctx, err, stack) {
     const causeStackStart = StringPrototypeIndexOf(causeStack, '\n    at');
     if (causeStackStart !== -1) {
       const causeFrames = StringPrototypeSplit(StringPrototypeSlice(causeStack, causeStackStart + 1), '\n');
-      const { len, offset } = identicalSequenceRange(frames, causeFrames);
+      const { 0: len, 1: offset } = identicalSequenceRange(frames, causeFrames);
       if (len > 0) {
         const skipped = len - 2;
         const msg = `    ... ${skipped} lines matching cause stack trace ...`;
@@ -1353,6 +1472,25 @@ function getStackFrames(ctx, err, stack) {
       }
     }
   }
+
+  // Remove recursive repetitive stack frames in long stacks
+  if (frames.length > 10) {
+    const ranges = getDuplicateErrorFrameRanges(frames);
+
+    for (let i = ranges.length - 3; i >= 0; i -= 3) {
+      const offset = ranges[i];
+      const length = ranges[i + 1];
+      const duplicateRanges = ranges[i + 2];
+
+      const msg = `    ... collapsed ${length * duplicateRanges} duplicate lines ` +
+        'matching above ' +
+        (duplicateRanges > 1 ?
+          `${length} lines ${duplicateRanges} times...` :
+          'lines ...');
+      frames.splice(offset, length * duplicateRanges, ctx.stylize(msg, 'undefined'));
+    }
+  }
+
   return frames;
 }
 
@@ -1412,16 +1550,45 @@ function removeDuplicateErrorKeys(ctx, keys, err, stack) {
 
 function markNodeModules(ctx, line) {
   let tempLine = '';
-  let nodeModule;
-  let pos = 0;
-  while ((nodeModule = nodeModulesRegExp.exec(line)) !== null) {
-    // '/node_modules/'.length === 14
-    tempLine += StringPrototypeSlice(line, pos, nodeModule.index + 14);
-    tempLine += ctx.stylize(nodeModule[1], 'module');
-    pos = nodeModule.index + nodeModule[0].length;
+  let lastPos = 0;
+  let searchFrom = 0;
+
+  while (true) {
+    const nodeModulePosition = StringPrototypeIndexOf(line, 'node_modules', searchFrom);
+    if (nodeModulePosition === -1) {
+      break;
+    }
+
+    // Ensure it's a path segment: must have a path separator before and after
+    const separator = line[nodeModulePosition - 1];
+    const after = line[nodeModulePosition + 12]; // 'node_modules'.length === 12
+
+    if ((after !== '/' && after !== '\\') || (separator !== '/' && separator !== '\\')) {
+      // Not a proper segment; continue searching
+      searchFrom = nodeModulePosition + 1;
+      continue;
+    }
+
+    const moduleStart = nodeModulePosition + 13; // Include trailing separator
+
+    // Append up to and including '/node_modules/'
+    tempLine += StringPrototypeSlice(line, lastPos, moduleStart);
+
+    let moduleEnd = StringPrototypeIndexOf(line, separator, moduleStart);
+    if (line[moduleStart] === '@') {
+      // Namespaced modules have an extra slash: @namespace/package
+      moduleEnd = StringPrototypeIndexOf(line, separator, moduleEnd + 1);
+    }
+
+    const nodeModule = StringPrototypeSlice(line, moduleStart, moduleEnd);
+    tempLine += ctx.stylize(nodeModule, 'module');
+
+    lastPos = moduleEnd;
+    searchFrom = moduleEnd;
   }
-  if (pos !== 0) {
-    line = tempLine + StringPrototypeSlice(line, pos);
+
+  if (lastPos !== 0) {
+    line = tempLine + StringPrototypeSlice(line, lastPos);
   }
   return line;
 }
@@ -1684,23 +1851,28 @@ function formatNumber(fn, number, numericSeparator) {
     }
     return fn(`${number}`, 'number');
   }
+
+  const numberString = String(number);
   const integer = MathTrunc(number);
-  const string = String(integer);
+
   if (integer === number) {
-    if (!NumberIsFinite(number) || StringPrototypeIncludes(string, 'e')) {
-      return fn(string, 'number');
+    if (!NumberIsFinite(number) || StringPrototypeIncludes(numberString, 'e')) {
+      return fn(numberString, 'number');
     }
-    return fn(`${addNumericSeparator(string)}`, 'number');
+    return fn(addNumericSeparator(numberString), 'number');
   }
   if (NumberIsNaN(number)) {
-    return fn(string, 'number');
+    return fn(numberString, 'number');
   }
+
+  const decimalIndex = StringPrototypeIndexOf(numberString, '.');
+  const integerPart = StringPrototypeSlice(numberString, 0, decimalIndex);
+  const fractionalPart = StringPrototypeSlice(numberString, decimalIndex + 1);
+
   return fn(`${
-    addNumericSeparator(string)
+    addNumericSeparator(integerPart)
   }.${
-    addNumericSeparatorEnd(
-      StringPrototypeSlice(String(number), string.length + 1),
-    )
+    addNumericSeparatorEnd(fractionalPart)
   }`, 'number');
 }
 
