@@ -7,8 +7,8 @@ const {
   AtomicsAdd,
   Float64Array,
   FunctionPrototypeBind,
-  JSONStringify,
   MathMax,
+  NumberMAX_SAFE_INTEGER,
   ObjectEntries,
   Promise,
   PromiseResolve,
@@ -19,6 +19,7 @@ const {
   String,
   StringPrototypeTrim,
   Symbol,
+  SymbolAsyncDispose,
   SymbolFor,
   TypedArrayPrototypeFill,
   Uint32Array,
@@ -40,6 +41,7 @@ const {
   ERR_WORKER_INVALID_EXEC_ARGV,
   ERR_INVALID_ARG_TYPE,
   ERR_INVALID_ARG_VALUE,
+  ERR_OPERATION_FAILED,
 } = errorCodes;
 
 const workerIo = require('internal/worker/io');
@@ -60,15 +62,17 @@ const { createMainThreadPort, destroyMainThreadPort } = require('internal/worker
 const { deserializeError } = require('internal/error_serdes');
 const { fileURLToPath, isURL, pathToFileURL } = require('internal/url');
 const { kEmptyObject } = require('internal/util');
-const { validateArray, validateString } = require('internal/validators');
+const { validateArray, validateString, validateObject, validateNumber } = require('internal/validators');
 const {
   throwIfBuildingSnapshot,
 } = require('internal/v8/startup_snapshot');
 const {
   ownsProcessState,
   isMainThread,
+  isInternalThread,
   resourceLimits: resourceLimitsRaw,
   threadId,
+  threadName,
   Worker: WorkerImpl,
   kMaxYoungGenerationSizeMb,
   kMaxOldGenerationSizeMb,
@@ -129,6 +133,37 @@ function assignEnvironmentData(data) {
   });
 }
 
+class CPUProfileHandle {
+  #worker = null;
+  #id = null;
+  #promise = null;
+
+  constructor(worker, id) {
+    this.#worker = worker;
+    this.#id = id;
+  }
+
+  stop() {
+    if (this.#promise) {
+      return this.#promise;
+    }
+    const stopTaker = this.#worker[kHandle]?.stopCpuProfile(this.#id);
+    return this.#promise = new Promise((resolve, reject) => {
+      if (!stopTaker) return reject(new ERR_WORKER_NOT_RUNNING());
+      stopTaker.ondone = (err, profile) => {
+        if (err) {
+          return reject(err);
+        }
+        resolve(profile);
+      };
+    });
+  };
+
+  async [SymbolAsyncDispose]() {
+    await this.stop();
+  }
+}
+
 class Worker extends EventEmitter {
   constructor(filename, options = kEmptyObject) {
     throwIfBuildingSnapshot('Creating workers');
@@ -165,8 +200,8 @@ class Worker extends EventEmitter {
       doEval = 'classic';
     } else if (isURL(filename) && filename.protocol === 'data:') {
       url = null;
-      doEval = 'module';
-      filename = `import ${JSONStringify(`${filename}`)}`;
+      doEval = 'data-url';
+      filename = `${filename}`;
     } else {
       doEval = false;
       if (isURL(filename)) {
@@ -203,7 +238,7 @@ class Worker extends EventEmitter {
         options.env);
     }
 
-    let name = '';
+    let name = 'WorkerThread';
     if (options.name) {
       validateString(options.name, 'options.name');
       name = StringPrototypeTrim(options.name);
@@ -339,9 +374,11 @@ class Worker extends EventEmitter {
       {
         const { stream, chunks } = message;
         const readable = this[kParentSideStdio][stream];
-        ArrayPrototypeForEach(chunks, ({ chunk, encoding }) => {
+        // This is a hot path, use a for(;;) loop
+        for (let i = 0; i < chunks.length; i++) {
+          const { chunk, encoding } = chunks[i];
           readable.push(chunk, encoding);
-        });
+        }
         return;
       }
       case messageTypes.STDIO_WANTS_MORE_DATA:
@@ -404,6 +441,10 @@ class Worker extends EventEmitter {
     });
   }
 
+  async [SymbolAsyncDispose]() {
+    await this.terminate();
+  }
+
   ref() {
     if (this[kHandle] === null) return;
 
@@ -424,6 +465,12 @@ class Worker extends EventEmitter {
     return this[kHandle].threadId;
   }
 
+  get threadName() {
+    if (this[kHandle] === null) return null;
+
+    return this[kHandle].threadName;
+  }
+
   get stdin() {
     return this[kParentSideStdio].stdin;
   }
@@ -456,6 +503,62 @@ class Worker extends EventEmitter {
       };
     });
   }
+
+  getHeapStatistics() {
+    const taker = this[kHandle]?.getHeapStatistics();
+
+    return new Promise((resolve, reject) => {
+      if (!taker) return reject(new ERR_WORKER_NOT_RUNNING());
+      taker.ondone = (handle) => {
+        resolve(handle);
+      };
+    });
+  }
+
+  cpuUsage(prev) {
+    if (prev) {
+      validateObject(prev, 'prev');
+      validateNumber(prev.user, 'prev.user', 0, NumberMAX_SAFE_INTEGER);
+      validateNumber(prev.system, 'prev.system', 0, NumberMAX_SAFE_INTEGER);
+    }
+    if (process.platform === 'sunos') {
+      throw new ERR_OPERATION_FAILED('worker.cpuUsage() is not available on SunOS');
+    }
+    const taker = this[kHandle]?.cpuUsage();
+    return new Promise((resolve, reject) => {
+      if (!taker) return reject(new ERR_WORKER_NOT_RUNNING());
+      taker.ondone = (err, current) => {
+        if (err !== null) {
+          return reject(err);
+        }
+        if (prev) {
+          resolve({
+            user: current.user - prev.user,
+            system: current.system - prev.system,
+          });
+        } else {
+          resolve({
+            user: current.user,
+            system: current.system,
+          });
+        }
+      };
+    });
+  }
+
+  // TODO(theanarkh): add options, such as sample_interval, CpuProfilingMode
+  startCpuProfile() {
+    const startTaker = this[kHandle]?.startCpuProfile();
+    return new Promise((resolve, reject) => {
+      if (!startTaker) return reject(new ERR_WORKER_NOT_RUNNING());
+      startTaker.ondone = (err, id) => {
+        if (err) {
+          return reject(err);
+        }
+        resolve(new CPUProfileHandle(this, id));
+      };
+    });
+  }
 }
 
 /**
@@ -538,6 +641,7 @@ module.exports = {
   ownsProcessState,
   kIsOnline,
   isMainThread,
+  isInternalThread,
   SHARE_ENV,
   resourceLimits:
     !isMainThread ? makeResourceLimits(resourceLimitsRaw) : {},
@@ -545,6 +649,7 @@ module.exports = {
   getEnvironmentData,
   assignEnvironmentData,
   threadId,
+  threadName,
   InternalWorker,
   Worker,
 };
