@@ -618,7 +618,14 @@ constexpr bool FastStringKey::operator==(const FastStringKey& other) const {
   return name_ == other.name_;
 }
 
-constexpr FastStringKey::FastStringKey(std::string_view name)
+consteval FastStringKey::FastStringKey(std::string_view name)
+    : FastStringKey(name, 0) {}
+
+constexpr FastStringKey FastStringKey::AllowDynamic(std::string_view name) {
+  return FastStringKey(name, 0);
+}
+
+constexpr FastStringKey::FastStringKey(std::string_view name, int dummy)
     : name_(name), cached_hash_(HashImpl(name)) {}
 
 constexpr std::string_view FastStringKey::as_string_view() const {
@@ -697,6 +704,31 @@ inline std::wstring ConvertToWideString(const std::string& str,
 }
 #endif  // _WIN32
 
+inline v8::MaybeLocal<v8::Object> NewDictionaryInstance(
+    v8::Local<v8::Context> context,
+    v8::Local<v8::DictionaryTemplate> tmpl,
+    v8::MemorySpan<v8::MaybeLocal<v8::Value>> property_values) {
+  for (auto& value : property_values) {
+    if (value.IsEmpty()) return v8::MaybeLocal<v8::Object>();
+  }
+  return tmpl->NewInstance(context, property_values);
+}
+
+inline v8::MaybeLocal<v8::Object> NewDictionaryInstanceNullProto(
+    v8::Local<v8::Context> context,
+    v8::Local<v8::DictionaryTemplate> tmpl,
+    v8::MemorySpan<v8::MaybeLocal<v8::Value>> property_values) {
+  for (auto& value : property_values) {
+    if (value.IsEmpty()) return v8::MaybeLocal<v8::Object>();
+  }
+  v8::Local<v8::Object> obj = tmpl->NewInstance(context, property_values);
+  if (obj->SetPrototypeV2(context, v8::Null(context->GetIsolate()))
+          .IsNothing()) {
+    return v8::MaybeLocal<v8::Object>();
+  }
+  return obj;
+}
+
 }  // namespace node
 
 #endif  // defined(NODE_WANT_INTERNALS) && NODE_WANT_INTERNALS
